from PySide2.QtCore    import QRegExp, Qt
from PySide2.QtGui     import QKeyEvent, QRegExpValidator
from PySide2.QtWidgets import QLineEdit
from qtrf.units        import prefix_map, regex_str, to_decimal

class LineEditMixin(object):
    def __init__(self):
        self.include_negative = True
        self.prefix_map       = prefix_map.copy()
        self.units            = 'U'
        self._special_keys = [Qt.Key_Space, Qt.Key_Enter, Qt.Key_Return]
        self.update_validation()

    def _update_prefix_keys(self):
        if self.prefix_map:
            self._prefix_keys = [ord(i.upper()) for i in self.prefix_map if i]
        else:
            self._prefix_keys = []
    def _create_validator(self):
        _regex_str = regex_str(self.prefix_map, self.units, self.include_negative)
        return QRegExpValidator(QRegExp(_regex_str))
    def _update_validator(self):
        self.setValidator(self._create_validator())
    def update_validation(self):
        self._update_prefix_keys()
        self._update_validator()

    def _is_special_key(self, key):
        return key in self._special_keys
    def _is_prefix_key(self, key):
        return key in self._prefix_keys
    def _prefix_value_from_ord(self, key):
        if not self.prefix_map:
            return None
        try:
            upper_chr = chr(key)
            lower_chr = upper_chr.lower()
        except ValueError:
            return None
        if upper_chr in self.prefix_map:
            return self.prefix_map[upper_chr]
        if lower_chr in self.prefix_map:
            return self.prefix_map[lower_chr]
        return None
    def keyPressEvent(self, event):
        key = event.key()
        if not self._is_special_key(key) and not self._is_prefix_key(key):
            QLineEdit.keyPressEvent(self, event)
            return
        prefix_value = self._prefix_value_from_ord(key)
        base_value   = to_decimal(self.text(), self.units, ignore_prefix=True)
        if prefix_value and base_value:
            self.value = base_value * prefix_value
        elif base_value:
            self.value = base_value
        self.selectAll()
        event.accept()

    def focusInEvent(self, event):
        QLineEdit.focusInEvent(self, event)
        self.selectAll()
    def focusOutEvent(self, event):
        self.value = self.value
        QLineEdit.focusOutEvent(self, event)



from .amps.label       import AmpsLabel
from .amps.line_edit   import AmpsLineEdit
from .db.label         import dBLabel
from .db.line_edit     import dBLineEdit
from .dbm.label        import dBmLabel
from .dbm.line_edit    import dBmLineEdit
from .file             import FileWidget
from .hz.label         import HzLabel
from .hz.line_edit     import HzLineEdit
from .meters.label     import MetersLabel
from .meters.line_edit import MetersLineEdit
from .ohms.label       import OhmsLabel
from .ohms.line_edit   import OhmsLineEdit
from .touchstone       import TouchstoneWidget
from .value.label      import ValueLabel
from .value.line_edit  import ValueLineEdit
from .volts.label      import VoltsLabel
from .volts.line_edit  import VoltsLineEdit



from decimal import Decimal, InvalidOperation, ROUND_HALF_UP

prefix_map = {
    'f': Decimal('1e-15'),
    'p': Decimal('1e-12'),
    'n': Decimal('1e-9'),
    'u': Decimal('1e-6'),
    'm': Decimal('1e-3'),
    '' : Decimal('1.0'),
    'K': Decimal('1e3'),
    'M': Decimal('1e6'),
    'G': Decimal('1e9'),
    'T': Decimal('1e12')
}

# validate str
def regex_str(prefixes=list(prefix_map), units='U', include_negative=True):
    sign_regex  = r'\-?' if include_negative else ''
    value_regex = r'[0-9]+(\.[0-9]*)?'
    units_regex = units or ''
    if prefixes:
        prefixes_regex = '[{}]?'.format(''.join(prefixes))
    else:
        prefixes_regex = ''
    return '{}{} {}{}'.format(sign_regex, value_regex, prefixes_regex, units).strip()

# str => number
def to_decimal(text, units='', ignore_prefix=False):
    # clean input
    try:
        text = text.strip().replace(' ', '')
    except AttributeError:
        return None
    if not text:
        return Decimal(1)
    # chop off units
    if units and text.endswith(units):
        unit_pos = -1*len(units)
        text = text[:unit_pos]
    # separate prefix
    prefix = ''
    if text[-1] in prefix_map:
        prefix = text[-1]
        text   = text[:-1]
    prefix = prefix if not ignore_prefix else ''
    try:
        return Decimal(text) * prefix_map[prefix]
    except InvalidOperation:
        if prefix and not ignore_prefix:
            return prefix_map[prefix]
        else:
            return None

# number => str
def to_base_and_prefix(value, prefix_map=prefix_map):
    if not prefix_map:
        return value, ''
    value     = Decimal(value)
    abs_value = abs  (value)
    prefix = list(prefix_map)[-1]
    for i in prefix_map:
        if abs_value < prefix_map[i] * Decimal('1e3'):
            prefix = i
            break
    return value/to_decimal(prefix), prefix
def round_decimal_places(value, decimal_places=3):
    value = Decimal(value)
    if decimal_places >= 1:
        zeros  = '0' * (decimal_places-1)
        format = Decimal('.' + zeros + '1')
    else:
        format = Decimal('1.')
    return value.quantize(format, ROUND_HALF_UP)
def to_str(value, decimal_places=3, prefix_map=prefix_map, units='U'):
    if not value:
        return ''
    base, prefix = to_base_and_prefix(value, prefix_map)
    base = round_decimal_places(base, decimal_places)
    return '{} {}{}'.format(base, prefix, units).strip()



from qtrf.widgets.value.line_edit import ValueLineEdit

class dBmLineEdit(ValueLineEdit):
    def __init__(self, parent=None):
        ValueLineEdit.__init__(self, parent)
        self.prefix_map = None
        self.units      = 'dBm'
        self.update_validation()



from decimal           import Decimal
from qtrf.units        import prefix_map, to_decimal, to_str

# Requires text(), setText(text) methods
# to exist in inheriting class
class ValueMixin(object):
    def __init__(self):
        self.decimal_places = 3
        self.prefix_map     = prefix_map.copy()
        self.units          = 'U'
    @property
    def value(self):
        return to_decimal(self.text(), self.units)
    @value.setter
    def value(self, value):
        if value:
            self.setText(self._to_str(value))
        else:
            self.clear()
    def _to_str(self, value):
        return to_str(value, self.decimal_places, self.prefix_map, self.units)



from PySide2.QtWidgets     import QLineEdit
from qtrf.mixins.line_edit import LineEditMixin
from qtrf.mixins.value     import ValueMixin

class ValueLineEdit(LineEditMixin, ValueMixin, QLineEdit):
    def __init__(self, parent=None):
        QLineEdit    .__init__(self, parent)
        ValueMixin   .__init__(self)
        LineEditMixin.__init__(self)
